from numpy import arange
from numpy.typing import NDArray
from scipy.interpolate import interp1d

from nested_mapping.nested_mapping import NestedMapping


def cross_check_refine_lsnl_data(
    storage: NestedMapping, *, xname: str, nominalname: str, **kwargs
) -> None:
    xcoarse = storage[xname]

    refiner = RefineGraph(xcoarse, **kwargs)

    for key, ycoarse in storage.walkitems():
        if ycoarse is xcoarse:
            continue

        nominal = storage[nominalname]
        storage[key] = refiner.process(ycoarse, nominal)

    storage[xname] = refiner.xfine


class RefineGraph:
    __slots__ = (
        "xcoarse",
        "xfine",
        "newmin",
        "newmax",
    )
    xcoarse: NDArray
    xfine: NDArray
    newmin: float
    newmax: float

    def __init__(
        self,
        xcoarse: NDArray,
        *,
        newmin: float,
        newmax: float,
    ):
        self.xcoarse = xcoarse
        self.newmin = newmin
        self.newmax = newmax

        self._process_x()

    def make_x(self) -> None:
        self.xfine = arange(0.0, 12.0000001, 0.05)

    def _process_x(self) -> None:
        self.make_x()

    def process(self, y: NDArray, nominal: NDArray) -> NDArray:
        skip_diff = y is nominal
        yfine = self._method_interpolate(y)
        yabs = self._method_reltoabs(yfine)

        return yabs if skip_diff else self._method_diff(nominal, yabs)

    def _method_reltoabs(self, yrel: NDArray) -> NDArray:
        return yrel * self.xfine

    def _method_interpolate(self, ycoarse) -> NDArray:
        fcn = interp1d(
            self.xcoarse,
            ycoarse,
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        return fcn(self.xfine)

    def _method_diff(self, nominal: NDArray, y: NDArray) -> NDArray:
        return nominal if nominal is y else y - nominal



from collections.abc import Sequence

from nested_mapping import NestedMapping


def refine_detector_data(
    source: NestedMapping,
    target: NestedMapping,
    *,
    detectors: Sequence[str],
    periods: Sequence[str] = ("6AD", "8AD", "7AD"),
    clean_source: bool = True,
    columns=("livetime", "eff", "efflivetime"),
    skip: Sequence[set[str]] | None = None,
) -> None:
    """Read arrays with detector data for the whole data taking period, write detector
    data for each period in separate array.
    """
    target["days"] = (days_storage := {})
    for det in detectors:
        day = source["day", det]
        step = day[1:] - day[:-1]
        assert (step == 1).all(), "Expect detector data for each day"

        ndet = source["ndet", det]
        for periodname in periods:
            period_ndet = int(periodname[0])
            mask_period = ndet == period_ndet

            for field in columns:
                data = source[field, det]

                key = (field, periodname, det)
                if skip is not None and any(skipkey.issubset(key) for skipkey in skip):
                    continue

                target[key] = data[mask_period]

            days = source["day", det][mask_period]
            days_stored = days_storage.setdefault(periodname, days)
            if days is not days_stored:
                assert all(days == days_stored)

    if clean_source:
        for key in tuple(source.walkkeys()):
            source.delete_with_parents(key)



from pathlib import Path
from numpy import ndarray, ascontiguousarray
from collections.abc import Sequence
from numpy.typing import NDArray


def validate_load_array(array: str | Path | Sequence[int | float] | NDArray | None) -> Path | NDArray | None:
    """Validate and load array.

    Parameters
    ----------
    array : str | Path | Sequence[int | float] | NDArray | None
        Path or array like object that will be validated. In case of array like
        object it will be converted to NDArray.

    Returns
    -------
    Path | NDArray | None
        NDArray object or validated path with description of array like object.
    """
    result = None
    match array:
        case str() | Path():
            result = Path(array)
            if not result.is_file():
                raise FileNotFoundError(f"File {result} is not found")
        case Sequence() | ndarray():
            result = ascontiguousarray(array, dtype="d")
        case None:
            result = None
        case _:
            raise RuntimeError(
                f"Invalid array type: {type(array).__name__}"
            )
    return result



from os import environ

from dag_modelling.tools.logger import set_verbosity
from numpy import allclose, fabs
from pytest import fixture, mark

from dgm_dayabay_dev.models import load_model

source_type_reference = "hdf5"
source_types_other = ["tsv", "npz", "root"]

precision_requirement = {"tsv": 2.0e-10, "root": 0, "npz": 0}
set_verbosity(1)


@fixture(scope="session")
def reference_model():
    return load_model(version="v1a", path_data=f"data/dayabay-v1a/{source_type_reference}")


@mark.parametrize("source_type", source_types_other)
def test_dayabay_source_type(reference_model, source_type: str):
    # TODO: automize test with `latest` version
    model = load_model(version="v1a", path_data=f"data/dayabay-v1a/{source_type}")

    outname = "outputs.eventscount.final.concatenated.detector_period"
    output_ref = reference_model.storage[outname]
    output = model.storage[outname]

    data_ref = output_ref.data
    data = output.data

    diff = fabs(data - data_ref)
    diff_rel = diff / data_ref
    diff_rel[data_ref == 0.0] = 0.0
    print(f"{source_type}: maxdiff {diff.max()} rel. diff max {diff_rel.max()}")

    rtol = precision_requirement[source_type]
    assert allclose(data, data_ref, rtol=rtol, atol=0.0), f"{source_type} requires {rtol=}"
